
/*
 * Catweasel -- Advanced Floppy Controller
 * Linux device driver
 * Low-level routines
 *
 * Copyright (C) 1998-2002 Michael Krause
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <asm/io.h>

#if defined(CONFIG_PCI) && LINUX_VERSION_CODE >= 0x020400
#include <linux/pci.h>
#endif

#include "catweasel.h"

static __inline__ void CWSetCReg(catweasel_contr *c, unsigned char clear, unsigned char set)
{
    c->control_register = (c->control_register & ~clear) | set;
    outb(c->control_register, c->io_sr);
}

static void CWTriggerStep(catweasel_contr *c)
{
    CWSetCReg(c, c->crm_step, 0);
    c->msdelay(5);
    CWSetCReg(c, 0, c->crm_step);
    c->msdelay(5);
}

void catweasel_init_controller(catweasel_contr *c)
{
    int i, j;

    if(!c->iobase)
	return;

    switch(c->type) {
    case CATWEASEL_TYPE_MK1:
	c->crm_sel0 = 1 << 5;
	c->crm_sel1 = 1 << 4;
	c->crm_mot0 = 1 << 3;
	c->crm_mot1 = 1 << 7;
	c->crm_dir  = 1 << 1;
	c->crm_step = 1 << 0;
	c->srm_trk0 = 1 << 4;
	c->srm_dchg = 1 << 5;
	c->srm_writ = 1 << 1;
	c->io_sr    = c->iobase + 2;
	c->io_mem   = c->iobase;
	break;
    case CATWEASEL_TYPE_MK3:
    case CATWEASEL_TYPE_MK4:
	c->crm_sel0 = 1 << 2;
	c->crm_sel1 = 1 << 3;
	c->crm_mot0 = 1 << 1;
	c->crm_mot1 = 1 << 5;
	c->crm_dir  = 1 << 4;
	c->crm_step = 1 << 7;
	c->srm_trk0 = 1 << 2;
	c->srm_dchg = 1 << 5;
	c->srm_writ = 1 << 6;
	c->io_sr    = c->iobase + 0xe8;
	c->io_mem   = c->iobase + 0xe0;
	break;
    default:
	return;	
    }

    c->control_register = 255;

    /* select all drives, step inside */
    CWSetCReg(c, c->crm_dir | c->crm_sel0 | c->crm_sel1, 0);
    for(i=0;i<3;i++)
	CWTriggerStep(c);

    for(i=0;i<2;i++) {
	c->drives[i].number = i;
	c->drives[i].contr = c;
	c->drives[i].diskindrive = 0;
	
	/* select only the respective drive, step to track 0 */
	if(i == 0) {
	    CWSetCReg(c, c->crm_sel0, c->crm_dir | c->crm_sel1);
	} else {
	    CWSetCReg(c, c->crm_sel1, c->crm_dir | c->crm_sel0);
	}
	
	for(j = 0; j < 86 && (inb(c->io_sr) & c->srm_trk0); j++)
	    CWTriggerStep(c);
	
	if(j < 86) {
	    c->drives[i].type = 1;
	    c->drives[i].track = 0;
	} else {
	    c->drives[i].type = 0;
	}
    }
    
    CWSetCReg(c, 0, c->crm_sel0 | c->crm_sel1); /* deselect all drives */
}

void catweasel_free_controller(catweasel_contr *c)
{
    if(!c->iobase)
	return;

    /* all motors off, deselect all drives */
    CWSetCReg(c, 0, c->crm_mot0 | c->crm_mot1 | c->crm_sel0 | c->crm_sel1);
}

void catweasel_select(catweasel_contr *c, int dr0, int dr1)
{
    CWSetCReg(c, dr0*c->crm_sel0 | dr1*c->crm_sel1, (!dr0)*c->crm_sel0 | (!dr1)*c->crm_sel1);
}

void catweasel_set_motor(catweasel_drive *d, int on)
{
    if(d->number == 0) {
	CWSetCReg(d->contr, on*d->contr->crm_mot0, (!on)*d->contr->crm_mot0);
    } else {
	CWSetCReg(d->contr, on*d->contr->crm_mot1, (!on)*d->contr->crm_mot1);
    }
}

int catweasel_seek(catweasel_drive *d, int t)
{
    int x;

    if(t == 0) {
	/* Recalibrate using explicit Track 0 sensing */
	catweasel_contr *c = d->contr;
	CWSetCReg(c, c->crm_dir, 0);
	for(x = 0; x < 3; x++)
	    CWTriggerStep(c);
	CWSetCReg(c, 0, c->crm_dir);

	for(x = 0; x < 86 && (inb(c->io_sr) & c->srm_trk0); x++)
	    CWTriggerStep(c);

	if(x == 86) {
	    return 0;
	}

	d->track = 0;
	return 1;
    }

    if(t >= 80) {
	return 0;
    }

    x = t - d->track;
    if(!x)
	return 1;

    if(x >= 0) {
	CWSetCReg(d->contr, d->contr->crm_dir, 0);
    } else {
	CWSetCReg(d->contr, 0, d->contr->crm_dir);
	x = -x;
    }

    while(x--)
	CWTriggerStep(d->contr);

    d->track = t;
    return 1;
}

int catweasel_disk_changed(catweasel_drive *d)
{
    int ot, t, changed = 0;

    if(inb(d->contr->io_sr) & d->contr->srm_dchg) {
	if(!d->diskindrive) {
	    /* first usage of this drive, issue disk change */
	    d->diskindrive = 1;
	    changed = 1;
	} else {
	    /* there's still a disk in there, no change detected. */
	}
    } else {
	ot = d->track;
	if(ot == 79)
	    t = 78;
	else
	    t = ot + 1;
	
	catweasel_seek(d, t);
	catweasel_seek(d, ot);
	
	if(!(inb(d->contr->io_sr) & d->contr->srm_dchg)) {
	    if(!d->diskindrive) {
		/* drive still empty, nothing has happened */
	    } else {
		/* disk has been removed */
		d->diskindrive = 0;
		changed = 1;
	    }
	} else {
	    /* disk has been inserted */
	    d->diskindrive = 1;
	    changed = 1;
	}
    }

    return changed;
}

int catweasel_write_protected(catweasel_drive *d)
{
    return !(inb(d->contr->io_sr) & 8);
}

int catweasel_read(catweasel_drive *d, int side, int clock, int time)
{
    int iobase = d->contr->iobase;

    if(!(inb(d->contr->io_sr) & d->contr->srm_dchg))
	return 0;

    if(d->contr->type == CATWEASEL_TYPE_MK1) {
	CWSetCReg(d->contr, 1<<2, (!side)<<2); /* set disk side */

	inb(iobase+1); /* ra reset */
	outb(clock*128, iobase+3);

	inb(iobase+1);
	inb(iobase+0);
	inb(iobase+0);
	outb(0, iobase+3); /* don't store index pulse */

	inb(iobase+1);

	inb(iobase+7); /* start reading */
	d->contr->msdelay(time); /* wait for one rotation */
	outb(0, iobase+1); /* stop reading, don't reset RAM pointer */

	outb(128, iobase+0); /* add data end mark */
	outb(128, iobase+0);

	inb(iobase+1); /* Reset RAM pointer */
    } else {
	CWSetCReg(d->contr, 1<<6, (!side)<<6); /* set disk side */

	outb(0, iobase + 0xe4); /* Reset memory pointer */
	switch(clock) {
	case 0: /* 28MHz */
	    outb(128, iobase + 0xec);
	    break;
	case 1: /* 14MHz */
	    outb(0, iobase + 0xec);
	    break;
	}
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	outb(0, iobase + 0xec); /* no IRQs, no MFM predecode */
	inb(iobase + 0xe0);
	outb(0, iobase + 0xec); /* don't store index pulse */

	outb(0, iobase + 0xe4); /* Reset memory pointer */
	inb(iobase + 0xf0); /* start reading */
	d->contr->msdelay(time); /* wait for one rotation */
	inb(iobase + 0xe4); /* stop reading, don't reset RAM pointer */

	outb(128, iobase + 0xe0); /* add data end mark */
	outb(0, iobase + 0xe4); /* Reset memory pointer */
    }

    return 1;
}

static void
catweasel_wait_for_writing_flag (catweasel_drive *d,
				 int status, /* Wait for this specified 'status' */
				 int timeout_ms)
{
    while(((inb(d->contr->io_sr) & d->contr->srm_writ) ? 1 : 0) != status && timeout_ms > 0) {
	d->contr->msdelay(20);
	timeout_ms -= 20;
    }
}

int
catweasel_write (catweasel_drive *d,
		 int side,
		 int clock,
		 int time)
{
    int iobase = d->contr->iobase;

    if(!(inb(d->contr->io_sr) & d->contr->srm_dchg))
	return 0;

    if(d->contr->type == CATWEASEL_TYPE_MK1) {
	CWSetCReg(d->contr, 1<<2, (!side)<<2); /* set disk side */

	inb(iobase+1);
	outb(clock*128, iobase+3);
	inb(iobase);
	outb(128, iobase+3); /* write enable */
	inb(iobase);
	inb(iobase);
	inb(iobase);
	inb(iobase);
	inb(iobase);
	inb(iobase);

	if(time > 0) {
	    outb(0, iobase+5);
	    d->contr->msdelay(time);
	} else {
	    outb(0, iobase+7);
	    catweasel_wait_for_writing_flag(d, 0, 300);
	    d->contr->msdelay(190);
	    catweasel_wait_for_writing_flag(d, 1, 50);
	}

	inb(iobase+1);
    } else {
	CWSetCReg(d->contr, 1<<6, (!side)<<6); /* set disk side */

	outb(0, iobase + 0xe4); /* Reset memory pointer */
	switch(clock) {
	case 0: /* 28MHz */
	    outb(128, iobase + 0xec);
	    break;
	case 1: /* 14MHz */
	    outb(0, iobase + 0xec);
	    break;
	}
	inb(iobase + 0xe0);
	outb(128, iobase + 0xec); /* write enable */
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);

	if(time > 0) {
	    outb(0, iobase + 0xf4); /* unconditional write START */
	    d->contr->msdelay(time);
	} else {
	    outb(0, iobase + 0xf0); /* ?????? indexed write START */
	    catweasel_wait_for_writing_flag(d, 0, 300);
	    d->contr->msdelay(190);
	    catweasel_wait_for_writing_flag(d, 1, 50);
	}

	inb(iobase + 0xe4); /* Abort writing */
	inb(iobase + 0xe4); /* Reset memory pointer */
    }

    return 1;
}
